import click

from tq42.cli.utils.types import TQ42CliContext, TQ42CliObject
from tq42.client import TQ42Client
from .auth_group import auth_group
from .organization_group import organization_group
from .experiment_group import experiment_group
from .project_group import project_group
from .environment_group import environment_group


@click.group()
@click.option(
    "--config",
    "config_path",
    required=False,
    help="Set path to alternate configuration file location",
)
@click.version_option()
@click.pass_context
def cli(ctx: TQ42CliContext, config_path: str):
    """
    Visit https://help.terraquantum.io/ to access our help center, from where you can access help articles and video tutorials, report bugs,
    contact support and request improvements.

    For TQ42SDK documentation, visit https://docs.tq42.com/en/latest/.

    Mandatory commands:

    tq42 auth login

    tq42 exp run create --exp="EXP_ID" --compute="COMPUTE_NAME" --algorithm="ALGORITHM_NAME" --parameters="PARAMETERS_JSON"

    ---

    All other command are optional.
    """
    ctx.ensure_object(TQ42CliObject)
    ctx.obj.client = TQ42Client(alt_config_file=config_path)


cli.add_command(auth_group)
cli.add_command(organization_group)
cli.add_command(experiment_group)
cli.add_command(project_group)
cli.add_command(environment_group)


if __name__ == "__main__":
    # TODO: do we need this? Maybe for testing?
    cli()



import unittest

from grpc import StatusCode
from grpc._channel import _InactiveRpcError as InactiveRpcError, _RPCState as RPCState

from tq42.utils.exceptions import (
    NoDefaultError,
    InvalidArgumentError,
    PermissionDeniedError,
    UnauthenticatedError,
    ExceedRetriesError,
)
from tq42.utils.exception_handling import handle_generic_sdk_errors
from tq42.utils.file_handling import read_file
from tq42.utils.constants import (
    no_default_error_file,
    unauthenticated_error_file,
    insufficient_permission_errors_file,
    invalid_arguments_error_file,
)


class TestTq42ErrorHandling(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def fail_if_still_here(self):
        self.assertTrue(False)

    def test_decorator_happy_path(self):
        @handle_generic_sdk_errors
        def happy_path():
            return "happy"

        self.assertEqual(happy_path(), "happy")

    def test_decorator_happy_path_with_params(self):
        @handle_generic_sdk_errors
        def summarize(a: int, b: int, c: int, d: int, e: int) -> int:
            return a + b + c + d + e

        self.assertEqual(summarize(1, 2, 3, 4, 5), 15)

    def test_decorator_invalid_argument_error(self):
        @handle_generic_sdk_errors
        def raise_invalid_argument():
            raise InvalidArgumentError("command", "details")

        try:
            raise_invalid_argument()
            self.fail_if_still_here()
        except InvalidArgumentError as e:
            self.assertEqual(
                str(e),
                read_file(invalid_arguments_error_file).format(e.command, e.details),
            )

    def test_decorator_grpc_invalid_argument_error(self):
        @handle_generic_sdk_errors
        def raise_invalid_argument_grpc():
            raise InactiveRpcError(
                RPCState(
                    code=StatusCode.INVALID_ARGUMENT,
                    details="no details",
                    due=[],
                    initial_metadata=[],
                    trailing_metadata=[],
                )
            )

        try:
            raise_invalid_argument_grpc()
            self.fail_if_still_here()
        except InvalidArgumentError as e:
            self.assertEqual(
                str(e),
                read_file(invalid_arguments_error_file).format(e.command, e.details),
            )

    def test_decorator_grpc_status_not_found_error(self):
        @handle_generic_sdk_errors
        def raise_status_not_found_grpc():
            raise InactiveRpcError(
                RPCState(
                    code=StatusCode.NOT_FOUND,
                    details="no details",
                    due=[],
                    initial_metadata=[],
                    trailing_metadata=[],
                )
            )

        try:
            raise_status_not_found_grpc()
            self.fail_if_still_here()
        except InvalidArgumentError as e:
            self.assertEqual(
                str(e),
                read_file(invalid_arguments_error_file).format(e.command, e.details),
            )

    def test_decorator_permission_denied_error(self):
        @handle_generic_sdk_errors
        def raise_permission_denied():
            raise PermissionDeniedError()

        try:
            raise_permission_denied()
            self.fail_if_still_here()
        except PermissionDeniedError as e:
            self.assertEqual(str(e), read_file(insufficient_permission_errors_file))

    def test_decorator_grpc_permission_denied_error(self):
        @handle_generic_sdk_errors
        def raise_permission_denied_grpc():
            raise InactiveRpcError(
                RPCState(
                    code=StatusCode.PERMISSION_DENIED,
                    details="no details",
                    due=[],
                    initial_metadata=[],
                    trailing_metadata=[],
                )
            )

        try:
            raise_permission_denied_grpc()
            self.fail_if_still_here()
        except PermissionDeniedError as e:
            self.assertEqual(str(e), read_file(insufficient_permission_errors_file))

    def test_decorator_unauthenticated_error(self):
        @handle_generic_sdk_errors
        def raise_unauthenticated():
            raise UnauthenticatedError()

        try:
            raise_unauthenticated()
            self.fail_if_still_here()
        except UnauthenticatedError as e:
            self.assertEqual(str(e), read_file(unauthenticated_error_file))

    def test_decorator_grpc_unauthenticated_error(self):
        @handle_generic_sdk_errors
        def raise_unauthenticated_grpc():
            raise InactiveRpcError(
                RPCState(
                    code=StatusCode.UNAUTHENTICATED,
                    details="no details",
                    due=[],
                    initial_metadata=[],
                    trailing_metadata=[],
                )
            )

        try:
            raise_unauthenticated_grpc()
            self.fail_if_still_here()
        except UnauthenticatedError as e:
            self.assertEqual(str(e), read_file(unauthenticated_error_file))

    def test_decorator_no_default_error(self):
        @handle_generic_sdk_errors
        def raise_no_default():
            raise NoDefaultError("command")

        try:
            raise_no_default()
            self.fail_if_still_here()
        except NoDefaultError as e:
            self.assertEqual(str(e), read_file(no_default_error_file).format(e.command))

    def test_decorator_key_error(self):
        @handle_generic_sdk_errors
        def raise_key_error():
            raise KeyError()

        try:
            raise_key_error()
            self.fail_if_still_here()
        except NoDefaultError as e:
            self.assertEqual(str(e), read_file(no_default_error_file).format(e.command))

    def test_decorator_exceed_retries_error(self):
        @handle_generic_sdk_errors
        def raise_exceed_retries():
            raise ExceedRetriesError(5)

        try:
            raise_exceed_retries()
            self.fail_if_still_here()
        except ExceedRetriesError as e:
            self.assertEqual(str(e), "Polling exceeded. Number of retries: 5")

    def test_generic_error(self):
        @handle_generic_sdk_errors
        def raise_generic_error():
            raise Exception("generic exception")

        self.assertRaises(Exception, raise_generic_error)



# disable the unused imports because we need them here for re-exporting purposes
# pylint: disable=unused-imports

# disable the * imports as we really want to re-export everything from these modules
# flake8: noqa: F403

# We import all algorithm related stuff from the grpc client here manually
# Unfortunately there currently is no way to dynamically import them AND have type hints.
# TODO: Revisit this to make it dynamic with working type hints

from com.terraquantum.experiment.v1.experimentrun.algorithm.shared_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.data_processing_shared_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.circuit_run_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_shared_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ts_mlp_train_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ts_mlp_eval_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ts_hqmlp_train_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ts_hqmlp_eval_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ts_lstm_train_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ts_lstm_eval_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ts_hqlstm_train_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ts_hqlstm_eval_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.cva_opt_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.tetra_opt_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.tetra_quenc_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.toy_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.generic_ml_train_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.generic_ml_infer_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.standard_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.classical_dense_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.classical_lstm_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.phn_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.pqn_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.efq_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.qdi_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.shared_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.qlstm_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.cphn_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.dhn_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.quantum_layer_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.custom_quantum_layer_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.gate_cnot_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.gate_encoding_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.gate_variational_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.gate_hadamard_pb2 import *
from com.terraquantum.experiment.v1.experimentrun.algorithm.ml_layers.gate_measurement_pb2 import *



from __future__ import annotations

import asyncio
import logging
from typing import List, Callable, Optional, Awaitable

import grpc.aio
from google.protobuf import empty_pb2

from tq42.utils.exception_handling import handle_generic_sdk_errors
from com.terraquantum.channel.v1alpha1.create_channel_pb2 import CreateChannelResponse

# important for re-export
from com.terraquantum.channel.v1alpha1.channel_message_pb2 import (
    ChannelMessage,
    Ask,
    Parameter,  # pylint:disable=unused-import # noqa: F401
    Tell,
    DataAcknowledge,
)

from typing import TYPE_CHECKING

# only import the stuff for type hints -> avoid circular imports
if TYPE_CHECKING:
    from tq42.client import TQ42Client


from tq42.utils.timers import AsyncTimedIterable

_RETRYABLE_ERROR_CODES = (
    grpc.StatusCode.DATA_LOSS,
    grpc.StatusCode.UNAVAILABLE,
    grpc.StatusCode.INTERNAL,
)
_MAX_RECONNECT_RETRIES = 5


class Channel:
    """
    Class to create and connect to a channel
    """

    id: str
    client: TQ42Client
    sequential_message_id: int = 0

    def __init__(self, client: TQ42Client, id: str):
        self.client = client
        self.id = id

    def __repr__(self) -> str:
        return f"<Channel Id={self.id}>"

    def __str__(self) -> str:
        return self.id

    @staticmethod
    @handle_generic_sdk_errors
    async def create(client: TQ42Client) -> Channel:
        empty = empty_pb2.Empty()
        res: CreateChannelResponse = await client.channel_client.CreateChannel(
            request=empty, metadata=client.metadata
        )
        return Channel(client=client, id=res.channel_id)

    async def connect(
        self,
        callback: Callable[[Ask], Awaitable[Tell]],
        finish_callback: Callable,
        max_duration_in_sec: Optional[int] = None,
        message_timeout_in_sec: Optional[int] = None,
    ) -> None:
        """
        Connects to the stream and handles every message with the provided callback to create an answer.
        ASK gets into the callback and then we expect a TELL answer

        :param callback: Async callback that handles an ASK message and returns a TELL message
        :param finish_callback: Callback that is called when channel is completed
        :param int max_duration_in_sec: Timeout for whole connection in seconds. `None` -> no timeout for overall flow
        :param int message_timeout_in_sec: Timeout between messages in seconds. `None` -> no timeout between messages
        """

        call = await self._establish_connection()

        async def _acknowledge_message(msg: ChannelMessage) -> None:
            nonlocal call
            ack_msg = ChannelMessage(
                acknowledge_data=DataAcknowledge(id=msg.sequential_message_id)
            )
            await call.write(ack_msg)
            logging.debug(f"User Sent ack {msg.sequential_message_id=}")

        async def _handle():
            nonlocal call
            try:
                timed_stream = AsyncTimedIterable(call, timeout=message_timeout_in_sec)
                incoming: ChannelMessage
                async for incoming in timed_stream:
                    logging.debug(f"User received {incoming=}")
                    data_field_name = incoming.WhichOneof("data")
                    if data_field_name == "completion_data":
                        await _acknowledge_message(msg=incoming)
                        logging.debug(
                            "Message indicated channel completion. Closing channel connection"
                        )
                        break
                    elif data_field_name == "ask_data":
                        if self.sequential_message_id >= incoming.sequential_message_id:
                            logging.debug(
                                "Message id is not sequential. Ignoring message"
                            )
                            continue
                        self.sequential_message_id = incoming.sequential_message_id
                        await _acknowledge_message(msg=incoming)
                        tell = await callback(incoming.ask_data)
                        tell_msg = ChannelMessage(
                            sequential_message_id=(incoming.sequential_message_id + 1),
                            tell_data=tell,
                        )
                        await call.write(tell_msg)

            except asyncio.TimeoutError:
                logging.debug("Stream finished because of the provided timeouts")
                raise TimeoutError(
                    f"Channel was closed due to exceeding {message_timeout_in_sec}s timeout between messages"
                ) from None

            except grpc.aio.AioRpcError as e:
                if e.code() in _RETRYABLE_ERROR_CODES:
                    call = await self._reestablish_connection()
                    if call:
                        return await _handle()

                raise e

        await asyncio.wait_for(_handle(), timeout=max_duration_in_sec)
        await call.done_writing()

        finish_callback()

    async def _establish_connection(self):
        metadata: tuple = (
            *self.client.metadata,
            ("channel-id", self.id),
        )
        call = self.client.channel_client.ConnectChannelCustomer(metadata=metadata)

        await call.write(ChannelMessage())
        return call

    async def _reestablish_connection(self):
        for i in range(1, _MAX_RECONNECT_RETRIES):
            logging.warning(
                f"Lost connection to channel, retrying (attempt {i}, max attempts: {_MAX_RECONNECT_RETRIES})"
            )
            try:
                await asyncio.wait_for(
                    self.client.channels_channel.channel_ready(), i * 2
                )
                call = await self._establish_connection()
                logging.info("Reconnected to channel.")
                return call
            except asyncio.TimeoutError:
                logging.error("Failed to reconnect.")

        return None


@handle_generic_sdk_errors
def list_all(client: TQ42Client) -> List[Channel]:
    """
    List all channels for a given user.

    For details, see (TODO: update link once a new documentation URL is created)
    """
    raise NotImplementedError("This is a functionality still to come")



from __future__ import annotations

from typing import Optional, List

from google.protobuf.field_mask_pb2 import FieldMask
from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf.json_format import MessageToJson

from tq42.utils.exception_handling import handle_generic_sdk_errors
from tq42.utils.pretty_list import PrettyList
from tq42.utils.cache import (
    get_current_value,
    write_key_value_to_cache,
    clear_cache,
)

from com.terraquantum.project.v1.project.project_pb2 import ProjectProto
from com.terraquantum.project.v1.project.get_project_request_pb2 import (
    GetProjectRequest,
)
from com.terraquantum.project.v1.project.update_project_request_pb2 import (
    UpdateProjectRequest,
)
from com.terraquantum.project.v1.project.list_projects_request_pb2 import (
    ListProjectsRequest,
)
from com.terraquantum.project.v1.project.list_projects_response_pb2 import (
    ListProjectsResponse,
)

from typing import TYPE_CHECKING

# only import the stuff for type hints -> avoid circular imports
if TYPE_CHECKING:
    from tq42.client import TQ42Client


class Project:
    """
    Class to manage projects

    https://docs.tq42.com/en/latest/Python_Developer_Guide/Setting_Up_Your_Environment.html#check-your-current-organization-and-project-settings
    """

    client: TQ42Client
    id: str
    data: ProjectProto

    def __init__(
        self, client: TQ42Client, id: str, data: Optional[ProjectProto] = None
    ) -> None:
        self.client = client
        self.id = id

        if data:
            self.data = data
        else:
            self.data = self._get()

    def __repr__(self) -> str:
        return f"<Project Id={self.id} Name={self.data.name}>"

    def __str__(self) -> str:
        return f"Project: {MessageToJson(self.data, preserving_proto_field_name=True)}"

    @handle_generic_sdk_errors
    def _get(self) -> ProjectProto:
        """
        Gets the data corresponding to this project id.
        """
        get_proj_request = GetProjectRequest(id=self.id)
        res = self.client.project_client.GetProject(
            request=get_proj_request, metadata=self.client.metadata
        )
        return res

    @staticmethod
    def from_proto(client: TQ42Client, msg: ProjectProto) -> Project:
        """
        Creates Project instance from a protobuf message.
        """
        return Project(client=client, id=msg.id, data=msg)

    @handle_generic_sdk_errors
    def update(self, name: str) -> Project:
        """
        Updates this specific project
        """
        project = {
            "id": self.id,
            "name": name,
        }

        # Create a new FieldMask instance
        field_mask = FieldMask()

        # Add paths to the FieldMask
        field_mask.paths.append("id")
        field_mask.paths.append("name")

        update_proj_request = UpdateProjectRequest(
            project=project,
            update_mask=field_mask,
            request_id=None,
        )
        self.data = self.client.project_client.UpdateProject(
            request=update_proj_request, metadata=self.client.metadata
        )
        return self

    @handle_generic_sdk_errors
    def set_friendly_name(self, friendly_name: str) -> Project:
        """
        Set a friendly name for a project.

        For details, see
        https://docs.tq42.com/en/latest/Python_Developer_Guide/Setting_Up_Your_Environment.html#setting-friendly-names-for-projects-and-experiments
        """
        return self.update(name=friendly_name)

    @staticmethod
    @handle_generic_sdk_errors
    def show(client: TQ42Client) -> Project:
        """
        Returns the current default project.
        If no default proj can be found return NoDefaultError

        https://docs.tq42.com/en/latest/Python_Developer_Guide/Setting_Up_Your_Environment.html#check-your-current-organization-and-project-settings
        """
        proj = get_current_value("proj")
        return Project(client=client, id=proj)

    @handle_generic_sdk_errors
    def set(self) -> Project:
        """
        If projectId is valid for the given user sets this projectId
        and the corresponding organizationId as the default

        https://docs.tq42.com/en/latest/Python_Developer_Guide/Setting_Up_Your_Environment.html#changing-your-workspace-to-a-different-organization-or-project
        """
        clear_cache()
        write_key_value_to_cache("org", self.data.organization_id)
        write_key_value_to_cache("proj", self.id)

        return self

    @staticmethod
    @handle_generic_sdk_errors
    def get_default(client: TQ42Client, organization_id: str) -> Optional[Project]:
        """
        Gets the default project for this user and organization_id based on the default_project field
        """
        project_list = list_all(client=client, organization_id=organization_id)
        if len(project_list) == 0:
            return None

        for proj in project_list:
            if proj.data.default_project and proj.data.default_project is True:
                return proj

        def access_created_at(project: Project) -> Timestamp:
            return project.data.created_at

        projects_sorted = sorted(project_list, key=access_created_at)
        return projects_sorted[0]


@handle_generic_sdk_errors
def list_all(
    client: TQ42Client, organization_id: Optional[str] = None
) -> List[Project]:
    """
    List all the projects you have permission to view within the organization that is currently set.

    For details, see
    https://docs.tq42.com/en/latest/Python_Developer_Guide/Setting_Up_Your_Environment.html#list-all-projects
    """

    if not organization_id:
        organization_id = get_current_value("org")

    create_list_proj_request = ListProjectsRequest(organization_id=organization_id)

    res: ListProjectsResponse = client.project_client.ListProjects(
        request=create_list_proj_request, metadata=client.metadata
    )
    return PrettyList(
        [Project.from_proto(client=client, msg=data) for data in res.projects]
    )



import click

import tq42.cli.utils.cli_functions as cli
from tq42.cli.utils.types import TQ42CliContext


@click.group("org")
def organization_group() -> click.Group:
    """
    Class to manage organization

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Setting_Up_Your_Environment.html#check-your-current-organization-and-project-settings
    """
    pass


@organization_group.command("list")
@click.pass_context
def list_all(ctx: TQ42CliContext) -> None:
    """
    List all the organizations you have permission to view.

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Setting_Up_Your_Environment.html#list-all-organizations
    """
    res = cli.list_orgs(ctx.obj.client)
    click.echo(res)


@organization_group.command("set")
@click.argument("org_id", required=True)
@click.pass_context
def org_set(ctx: TQ42CliContext, org_id: str) -> None:
    """
    Change the active organization.

    e.g.: tq42 org set 5bac0b60-48d0-45cd-bf0a-39505b058106

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Setting_Up_Your_Environment.html#changing-your-workspace-to-a-different-organization-or-project
    """
    res = cli.set_org(ctx.obj.client, org_id)
    click.echo(res)



from dataclasses import dataclass
from typing import Any

from click.testing import CliRunner

from tq42.cli.entrypoint import cli
from tq42.client import TQ42Client
from tq42.organization import list_all as list_all_organizations
from tq42.project import list_all as list_all_projects
from tq42.experiment import list_all as list_all_experiments
from tq42.experiment_run import list_all as list_all_experiment_runs


@dataclass()
class Args:
    org: str
    proj: str
    exp: str
    run: str
    export_path: str
    config: Any


class FunctionalTestConfig:
    args: Args = None
    _client = None

    @property
    def org(self):
        return self.args.org

    @property
    def proj(self):
        return self.args.proj

    @property
    def exp(self):
        return self.args.exp

    @property
    def exp_run(self):
        return self.args.run

    @property
    def export_path(self):
        return self.args.export_path

    @property
    def config(self):
        return self.args.config

    def get_client(self) -> TQ42Client:
        if self._client is None:
            self._client = TQ42Client(self.config)
        return self._client

    @staticmethod
    def prepare_defaults():
        auto_pick = False
        client = TQ42Client()
        arguments = Args(org="", proj="", exp="", run="", export_path="", config=None)

        choices = [
            "{} ({})".format(org.id, org.data.name)
            for org in list_all_organizations(client=client)
        ]
        arguments.org = FunctionalTestConfig.provide_choice_id(
            "org", choices, auto_pick
        )

        choices = [
            "{} ({})".format(proj.id, proj.data.name)
            for proj in list_all_projects(client=client, organization_id=arguments.org)
        ]
        arguments.proj = FunctionalTestConfig.provide_choice_id(
            "proj", choices, auto_pick
        )
        choices = [
            "{} ({})".format(exp.id, exp.data.name)
            for exp in list_all_experiments(client=client, project_id=arguments.proj)
        ]
        arguments.exp = FunctionalTestConfig.provide_choice_id(
            "exp", choices, auto_pick
        )
        choices = [
            "{}".format(runs.id)
            for runs in list_all_experiment_runs(
                client=client, experiment_id=arguments.exp
            )
        ]
        arguments.run = FunctionalTestConfig.provide_choice_id(
            "exp run", choices, auto_pick
        )
        arguments.export_path = FunctionalTestConfig.question_with_default(
            "Which export path to use?", ".", auto_pick
        )
        return arguments

    @staticmethod
    def question_with_default(question: str, default: str, auto_pick: bool) -> str:
        if auto_pick:
            return default

        choice = str(input(f"{question} ({default})"))
        if choice.strip() == "":
            return default

        return choice

    @staticmethod
    def custom_select(question: str, choices: list) -> str:
        print(question)
        for index, choice in enumerate(choices, start=1):
            print(f"{index}. {choice}")

        while True:
            try:
                choice_index = int(input("Enter the number of your choice: ")) - 1
                if 0 <= choice_index < len(choices):
                    return choices[choice_index]
                else:
                    print("Invalid choice. Please enter a valid number.")
            except ValueError:
                print("Invalid input. Please enter a number.")

    @staticmethod
    def provide_choice_id(arg_type: str, choices: [str], auto_pick: bool) -> str:
        if auto_pick:
            return choices[0].split(" ")[0]

        choice = FunctionalTestConfig.custom_select(
            "Which {} do you want to use?".format(arg_type), choices=choices
        )
        return choice.split(" ")[0]


class FunctionalCLITestConfig(FunctionalTestConfig):
    runner = CliRunner()
    cli_entry = cli

    @property
    def client(self):
        return self.get_client()



from __future__ import annotations

import base64
import hashlib
import os.path
from pathlib import Path
from typing import Optional, List
from urllib.error import HTTPError

from google.protobuf.json_format import MessageToJson
from tqdm import tqdm
import requests
import validators

from tq42.client import TQ42Client
from tq42.utils.exception_handling import handle_generic_sdk_errors

# This is important to re-export it!
from com.terraquantum.storage.v1alpha1.storage_pb2 import (
    DatasetSensitivityProto,
    StorageProto,
    StorageType,
)
from com.terraquantum.storage.v1alpha1.create_storage_from_file_pb2 import (
    CreateStorageFromFileRequest,
    CreateStorageFromFileResponse,
)
from com.terraquantum.storage.v1alpha1.create_storage_from_external_pb2 import (
    CreateStorageFromExternalBucketRequest,
)
from com.terraquantum.storage.v1alpha1.get_storage_request_pb2 import GetStorageRequest
from com.terraquantum.storage.v1alpha1.list_storages_pb2 import (
    ListStoragesRequest,
    ListStoragesResponse,
)
from com.terraquantum.storage.v1alpha1.export_storage_pb2 import (
    ExportStorageRequest,
    ExportStorageResponse,
)

from tq42.utils.pretty_list import PrettyList


class Dataset:
    """
    Class to create and view datasets

    https://docs.tq42.com/en/latest/Python_Developer_Guide/Work_with_Datasets.html
    """

    id: str
    data: StorageProto
    client: TQ42Client

    def __init__(
        self, client: TQ42Client, id: str, data: Optional[StorageProto] = None
    ):
        self.client = client
        self.id = id

        if data:
            self.data = data
        else:
            self.data = self._get()

    def __repr__(self) -> str:
        return f"<Dataset Id={self.id} Name={self.data.name}>"

    def __str__(self) -> str:
        return f"Dataset: {MessageToJson(self.data, preserving_proto_field_name=True)}"

    @handle_generic_sdk_errors
    def _get(self) -> StorageProto:
        get_storage_request = GetStorageRequest(storage_id=self.id)
        storage_data: StorageProto = self.client.storage_client.GetStorage(
            request=get_storage_request, metadata=self.client.metadata
        )
        return storage_data

    def _refresh(self) -> None:
        self.data = self._get()

    @staticmethod
    def from_proto(client: TQ42Client, msg: StorageProto) -> Dataset:
        """
        Creates Dataset instance from a protobuf message.
        """
        return Dataset(client=client, id=msg.id, data=msg)

    @staticmethod
    @handle_generic_sdk_errors
    def create(
        client: TQ42Client,
        project_id: str,
        name: str,
        description: str,
        sensitivity: DatasetSensitivityProto,
        file: str = None,
        url: str = None,
    ) -> Dataset:
        """
        Create a dataset for a project.

        For details, see https://docs.tq42.com/en/latest/Python_Developer_Guide/Working_with_Datasets.html
        """

        if (file and url) or (not file and not url):
            raise ValueError("Please provide (only) one of: file or url")

        if url:
            res = Dataset._create_from_external_bucket(
                client=client,
                project_id=project_id,
                name=name,
                description=description,
                url=url,
                sensitivity=sensitivity,
            )
            return Dataset.from_proto(client=client, msg=res)

        res = Dataset._create_from_file(
            client=client,
            project_id=project_id,
            name=name,
            description=description,
            file=file,
            sensitivity=sensitivity,
        )
        return Dataset.from_proto(client=client, msg=res)

    @staticmethod
    def _create_from_file(
        client: TQ42Client,
        project_id: str,
        name: str,
        description: str,
        file: str,
        sensitivity: DatasetSensitivityProto,
    ) -> StorageProto:
        file_path = Path(file)
        if not file_path.exists():
            raise FileNotFoundError("The specified file does not exist")

        with file_path.open(mode="rb") as f:
            data = f.read()
            file_hash = hashlib.md5(data).digest()
            file_hash_b64 = base64.b64encode(file_hash).decode("utf-8")

            create_dataset_request = CreateStorageFromFileRequest(
                project_id=project_id,
                name=name,
                description=description,
                hash_md5=file_hash_b64,
                file_name=file_path.name,
                sensitivity=sensitivity,
            )

            res: CreateStorageFromFileResponse = (
                client.storage_client.CreateStorageFromFile(
                    request=create_dataset_request, metadata=client.metadata
                )
            )

            headers = {
                "Content-Type": "application/octet-stream",
                "Content-MD5": file_hash_b64,
            }
            file_upload_response = requests.put(
                url=res.signed_url,
                headers=headers,
                data=data,
            )

            if not file_upload_response.ok:
                raise HTTPError(
                    url=res.signed_url,
                    code=file_upload_response.status_code,
                    msg=f"Upload of file {file} to storage failed. Please make sure your network is working. "
                    "If issues persist please get in touch via https://help.terraquantum.io/en",
                    fp=None,
                    hdrs=file_upload_response.headers,
                )

        return res.storage

    @staticmethod
    def _create_from_external_bucket(
        client: TQ42Client,
        project_id: str,
        name: str,
        description: str,
        url: str,
        sensitivity: DatasetSensitivityProto,
    ) -> StorageProto:
        create_dataset_request = CreateStorageFromExternalBucketRequest(
            project_id=project_id,
            name=name,
            description=description,
            url=url,
            sensitivity=sensitivity,
        )

        return client.storage_client.CreateStorageFromExternalBucket(
            request=create_dataset_request, metadata=client.metadata
        )

    @handle_generic_sdk_errors
    def export(self, directory_path: str) -> List[str]:
        """
        Export all files within the dataset.
        Returns a list of exported file paths

        For details, see https://docs.tq42.com/en/latest/Python_Developer_Guide/Working_with_Datasets.html
        """
        if not os.path.isdir(directory_path):
            raise ValueError(
                f"Provided directory path {directory_path} is not a valid directory"
            )

        export_storage_request = ExportStorageRequest(storage_id=self.id)

        res: ExportStorageResponse = self.client.storage_client.ExportStorage(
            request=export_storage_request, metadata=self.client.metadata
        )

        exported_file_paths = []

        for signed_url in res.signed_urls:
            file_path = os.path.join(
                directory_path,
                self._get_file_name_from_signed_url(signed_url=signed_url),
            )
            self._download_file_from_url(url=signed_url, file_path=file_path)
            exported_file_paths.append(file_path)

        return exported_file_paths

    @staticmethod
    def _download_file_from_url(url: str, file_path: str):
        response = requests.get(url, stream=True)

        if os.path.exists(file_path):
            raise FileExistsError(file_path)

        print("Downloading file to {}".format(file_path))
        with open(file_path, "wb") as handle:
            for data in tqdm(response.iter_content()):
                handle.write(data)

    @staticmethod
    def _get_file_name_from_signed_url(signed_url: str) -> str:
        if not validators.url(signed_url):
            raise ValueError(f"The signed URL {signed_url} is not a valid URL")
        url_without_parameters = signed_url.split("?")[0]
        return url_without_parameters.split("/")[-1]


@handle_generic_sdk_errors
def list_all(client: TQ42Client, project_id: str) -> List[Dataset]:
    """
    List all datasets for a project.

    For details, see https://docs.tq42.com/en/latest/Python_Developer_Guide/Working_with_Datasets.html
    """
    list_datasets_request = ListStoragesRequest(
        project_id=project_id, type=StorageType.DATASET
    )
    res: ListStoragesResponse = client.storage_client.ListStorages(
        request=list_datasets_request, metadata=client.metadata
    )
    return PrettyList(
        [Dataset.from_proto(client=client, msg=dataset) for dataset in res.storages]
    )



import click

from tq42.cli.utils.types import TQ42CliContext
from tq42.utils.environment import (
    get_environment,
    environment_default_set,
    environment_clear,
)


@click.group("env")
def environment_group() -> click.Group:
    pass


@environment_group.command("print")
def print_environment():
    click.echo(get_environment())


@environment_group.command("default")
@click.pass_context
def default_environment(ctx: TQ42CliContext):
    click.echo(environment_default_set(client=ctx.obj.client))


@environment_group.command("clear")
def clear_environment():
    click.echo(environment_clear())



from __future__ import annotations

from typing import Optional, List

from google.protobuf.json_format import MessageToJson

from tq42.client import TQ42Client
from tq42.utils.exception_handling import handle_generic_sdk_errors

from com.terraquantum.storage.v1alpha1.storage_pb2 import (
    StorageProto,
    StorageType,
)
from com.terraquantum.storage.v1alpha1.get_storage_request_pb2 import GetStorageRequest
from com.terraquantum.storage.v1alpha1.list_storages_pb2 import (
    ListStoragesRequest,
    ListStoragesResponse,
)

from tq42.utils.pretty_list import PrettyList


class Model:
    """
    Class to view models

    https://docs.tq42.com/en/latest/Python_Developer_Guide/Working_with_Models.html
    """

    id: str
    data: StorageProto
    client: TQ42Client

    def __init__(
        self, client: TQ42Client, id: str, data: Optional[StorageProto] = None
    ):
        self.client = client
        self.id = id

        if data:
            self.data = data
        else:
            self.data = self._get()

    def __repr__(self) -> str:
        return f"<Model Id={self.id} Name={self.data.name}>"

    def __str__(self) -> str:
        return f"Model: {MessageToJson(self.data, preserving_proto_field_name=True)}"

    @handle_generic_sdk_errors
    def _get(self) -> StorageProto:
        get_storage_request = GetStorageRequest(storage_id=self.id)
        storage_data: StorageProto = self.client.storage_client.GetStorage(
            request=get_storage_request, metadata=self.client.metadata
        )
        return storage_data

    def _refresh(self) -> None:
        self.data = self._get()

    @staticmethod
    def from_proto(client: TQ42Client, msg: StorageProto) -> Model:
        """
        Creates model instance from a protobuf message.
        """
        return Model(client=client, id=msg.id, data=msg)


@handle_generic_sdk_errors
def list_all(client: TQ42Client, project_id: str) -> List[Model]:
    """
    List all models for a project.

    For details, see https://docs.tq42.com/en/latest/Python_Developer_Guide/Working_with_Models.html
    """
    list_models_request = ListStoragesRequest(
        project_id=project_id, type=StorageType.MODEL
    )
    res: ListStoragesResponse = client.storage_client.ListStorages(
        request=list_models_request, metadata=client.metadata
    )
    return PrettyList(
        [Model.from_proto(client=client, msg=model) for model in res.storages]
    )



import unittest

from tq42.functional_tests.functional_test_config import FunctionalCLITestConfig


class TestFunctionalInvokeCliProjectGroup(unittest.TestCase, FunctionalCLITestConfig):
    def test_proj_show(self):
        result = self.runner.invoke(self.cli_entry, ["proj", "show"])
        assert result.exit_code == 0
        assert result.output is not None

    def test_proj_list(self):
        result = self.runner.invoke(self.cli_entry, ["proj", "list"])
        assert result.exit_code == 0
        assert result.output is not None

    def test_proj_list_with_org_parameter(self):
        result = self.runner.invoke(
            self.cli_entry, ["proj", "list", "--org=" + self.org]
        )
        assert result.exit_code == 0
        assert result.output is not None

    def test_proj_set(self):
        result = self.runner.invoke(self.cli_entry, ["proj", "set", self.proj])
        assert result.exit_code == 0
        assert result.output is not None

    def test_proj_set_friendly_name(self):
        result = self.runner.invoke(
            self.cli_entry,
            ["proj", "set-friendly-name", "FRIENDLY_NAME", "--proj=" + self.proj],
        )
        assert result.exit_code == 0
        assert result.output is not None



import unittest
import uuid

from tq42.organization import list_all as list_all_organizations, Organization
from tq42.project import Project, list_all as list_all_projects
from tq42.experiment import Experiment, list_all as list_all_experiments
from tq42.experiment_run import (
    ExperimentRun,
    list_all as list_all_experiment_runs,
    HardwareProto,
)
from tq42.functional_tests.functional_test_config import FunctionalTestConfig
from tq42.algorithm import (
    ToyMetadataProto,
    ToyParametersProto,
    ToyInputsProto,
    AlgorithmProto,
)
from google.protobuf.json_format import MessageToDict

from com.terraquantum.experiment.v1.experimentrun.experiment_run_pb2 import (
    ExperimentRunStatusProto,
)


class TestFunctionalTQ42API(unittest.TestCase, FunctionalTestConfig):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_org_list(self):
        org_list_obj = list_all_organizations(client=self.get_client())

        # Expect that there is at least one project in project list
        self.assertGreater(len(org_list_obj), 0)

        # Expect that specific project is in project list
        org_ids = [org.id for org in org_list_obj]
        self.assertIn(self.org, org_ids)

    def test_org_get(self):
        org = Organization(client=self.get_client(), id=self.org)
        self.assertEqual(self.org, org.id)
        # TODO: We need the appropriate enum for status
        self.assertEqual(1, org.data.state)
        self.assertTrue(org.data.default_org)

    def test_proj_list(self):
        proj_list = list_all_projects(
            client=self.get_client(), organization_id=self.org
        )

        # Expect that there is at least one project in project list
        self.assertGreater(len(proj_list), 0)

        # Expect that specific project is in project list
        proj_ids = [proj.id for proj in proj_list]
        self.assertIn(self.proj, proj_ids)

    def test_proj_get(self):
        proj = Project(client=self.get_client(), id=self.proj)

        self.assertEqual(self.proj, proj.id)
        self.assertEqual(self.org, proj.data.organization_id)

    def test_exp_list(self):
        exp_list = list_all_experiments(client=self.get_client(), project_id=self.proj)

        # Expect that there is at least one experiment in experiment list
        self.assertGreater(len(exp_list), 0)

        # Expect that specific experiment is in experiment list
        exp_ids = [exp.id for exp in exp_list]
        self.assertIn(self.exp, exp_ids)

    def test_exp_get(self):
        exp = Experiment(client=self.get_client(), id=self.exp)

        self.assertEqual(self.exp, exp.id)
        self.assertEqual(self.proj, exp.data.project_id)

    def test_exp_run_list(self):
        exp_runs = list_all_experiment_runs(
            client=self.get_client(), experiment_id=self.exp
        )

        # Expect that there is at least one experiment run for this experiment
        self.assertGreater(len(exp_runs), 0)

        # Expect that specific experiment run is in experiment list
        exp_run_ids = [exp_run.id for exp_run in exp_runs]
        self.assertIn(self.exp_run, exp_run_ids)

    def test_exp_run_get(self):
        exp_run = ExperimentRun(client=self.get_client(), id=self.exp_run)

        self.assertEqual(self.exp_run, exp_run.id)
        self.assertEqual(self.exp, exp_run.data.experiment_id)

    def test_exp_run_create(self):
        parameters = ToyMetadataProto(
            parameters=ToyParametersProto(n=2, r=1, msg="correct"),
            inputs=ToyInputsProto(),
        )
        parameters = MessageToDict(parameters, preserving_proto_field_name=True)
        exp_run = ExperimentRun.create(
            client=self.get_client(),
            algorithm=AlgorithmProto.TOY,
            experiment_id=self.exp,
            compute=HardwareProto.SMALL,
            parameters=parameters,
        )
        self.assertEqual(self.exp, exp_run.data.experiment_id)
        self.assertEqual(ExperimentRunStatusProto.QUEUED, exp_run.data.status)

    def test_exp_set_friendly_name(self):
        random_string = str(uuid.uuid4())
        updated_exp = Experiment(client=self.get_client(), id=self.exp).update(
            name=random_string
        )
        self.assertEqual(updated_exp.data.name, random_string)

    def test_proj_set_friendly_name(self):
        random_string = str(uuid.uuid4())
        updated_project = Project(client=self.get_client(), id=self.proj).update(
            name=random_string
        )
        self.assertEqual(updated_project.data.name, random_string)



import ast

import click

import tq42.cli.utils.cli_functions as cli
from tq42.cli.utils.types import TQ42CliContext
from tq42.experiment_run import ExperimentRun, HardwareProto
from tq42.algorithm import AlgorithmProto


@click.group("run")
def exp_run_group() -> click.Group:
    """
    Class to run experiments and view results

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Submitting_and_Monitoring_a_Run.html#
    """
    pass


@exp_run_group.command("cancel")
@click.argument("exp_run_id", required=True, type=str)
@click.pass_context
def cancel_run(ctx: TQ42CliContext, exp_run_id: str) -> None:
    """
    Cancel a run that is QUEUED, PENDING, or RUNNING.

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Submitting_and_Monitoring_a_Run.html#cancelling-an-experiment-run
    """
    res = cli.cancel_exprun(ctx.obj.client, exp_run_id)
    click.echo(res)


@exp_run_group.command("poll")
@click.argument("exp_run_id", required=True, type=str)
@click.pass_context
def poll_run(ctx: TQ42CliContext, exp_run_id: str) -> None:
    """
    Monitor an experiment run until it completes, then automatically display the results (if there are no errors).

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Submitting_and_Monitoring_a_Run.html#monitoring-an-experiment-run
    """
    res = cli.poll_exprun(ctx.obj.client, exp_run_id)
    click.echo(res)


@exp_run_group.command("check")
@click.argument("exp_run_id", required=True, type=str)
@click.pass_context
def check_run(ctx: TQ42CliContext, exp_run_id: str) -> None:
    """
    Monitor run status.

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Submitting_and_Monitoring_a_Run.html#monitoring-an-experiment-run
    """
    res = cli.get_exprun(ctx.obj.client, exp_run_id)
    click.echo(res)


@exp_run_group.command("list")
@click.option(
    "--exp",
    "exp_id",
    required=True,
    type=str,
    help="Attribute the command to a particular experiment",
)
@click.pass_context
def list_runs(ctx: TQ42CliContext, exp_id: str) -> None:
    """
    List all the runs within an experiment you have permission to view.

    e.g. tq42 exp run list --exp 98ccb1d2-a3d0-48c8-b172-022f6db9be01

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Setting_Up_Your_Environment.html#list-all-runs-within-an-experiment
    """
    res = cli.list_expruns(ctx.obj.client, exp_id)
    click.echo(res)


@exp_run_group.command("create")
@click.option(
    "--exp",
    "exp_id",
    required=True,
    type=str,
    help="Attribute the command to a particular experiment",
)
@click.option(
    "--compute",
    "compute",
    required=True,
    type=str,
    help="Specify the compute resources to apply to the command",
)
@click.option(
    "--algorithm",
    "algorithm",
    required=True,
    type=str,
    help="Specify the algorithm to apply to the command",
)
@click.option(
    "--parameters",
    "parameters",
    required=True,
    type=str,
    help="Specify the parameters to apply to the command",
)
@click.pass_context
def create_run(
    ctx: TQ42CliContext, exp_id: str, compute: str, algorithm: str, parameters: str
) -> None:
    """
    Begin an experiment run.

    e.g. tq42 exp run create --exp 98ccb1d2-a3d0-48c8-b172-022f6db9be01  --compute small --algorithm TETRA_OPT --parameters \"{'parameters': {'dimensionality':6,'maximal_rank' :1, 'points_number': 1, 'quantization' : True , 'tolerance':3.9997,  'grid': [1,2,3], 'upper_limits':[1,2,3,4,6,6], 'lower_limits': [0,0,0,0,0,0] , 'objective_function':'https://terraquantum.swiss', 'iteration_number': 1}, 'inputs': {}}\"

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Submitting_and_Monitoring_a_Run.html#submitting-an-experiment-run
    """
    algo = AlgorithmProto.Value(algorithm.upper())
    compute_val = HardwareProto.Value(compute.upper())
    params = ast.literal_eval(parameters)
    click.echo(
        ExperimentRun.create(
            client=ctx.obj.client,
            algorithm=algo,
            experiment_id=exp_id,
            compute=compute_val,
            parameters=params,
        ).data
    )



import unittest
from dataclasses import dataclass

from tq42.cli.utils import formatter


class MockWithID:
    def __init__(self, id):
        self.id = id


@dataclass
class NameData:
    name: str


class MockWithNameResourceClasses(MockWithID):
    def __init__(self, id, name=None):
        super().__init__(id)
        self.data = NameData(name)


class MockWithName(MockWithID):
    def __init__(self, id, name=None):
        super().__init__(id)
        self.name = name


class MockList:
    def __init__(self, a):
        self.a = a


class MockOrgList(MockList):
    @property
    def organizations(self):
        return self.a


class MockProjList(MockList):
    @property
    def projects(self):
        return self.a


class MockExpList(MockList):
    @property
    def experiments(self):
        return self.a


class TestOutputFormat(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_format_org(self):
        org = MockWithID("ORG_ID")
        self.assertEqual('org="ORG_ID"', formatter.org_formatter.format(org))

    def test_format_org_by_project(self):
        proj = MockWithNameResourceClasses("PROJ_ID", name="Cool project")
        proj.data.organization_id = "ORG_ID"
        self.assertEqual(
            'org="ORG_ID"', formatter.org_formatter.format_by_project(proj)
        )

    def test_set_proj_lines(self):
        proj = MockWithNameResourceClasses("PROJ_ID", name="Cool project")
        proj.data.organization_id = "ORG_ID"
        expected = ['org="ORG_ID"', 'proj="Cool project" (PROJ_ID)']
        actual = formatter.set_proj_lines(proj)
        self.assertEqual(expected, actual)

    def test_format_org_list(self):
        a = [MockWithID("ORG_ID{}".format(i)) for i in range(2)]
        expected = ['org="ORG_ID0"', 'org="ORG_ID1"']
        actual = formatter.org_formatter.format_by_list_object(a)
        self.assertEqual(expected, actual)

    def test_format_proj(self):
        no_name_proj = MockWithNameResourceClasses("PROJ_ID0")
        proj = MockWithNameResourceClasses("PROJ_ID1", name="Cool Project!")
        self.assertEqual(
            'proj="PROJ_ID0"', formatter.proj_formatter.format(no_name_proj)
        )
        self.assertEqual(
            'proj="Cool Project!" (PROJ_ID1)', formatter.proj_formatter.format(proj)
        )

    def test_format_proj_list(self):
        a = [MockWithNameResourceClasses("PROJ_ID{}".format(i)) for i in range(2)]
        a[1].data.name = "Cool Project!"
        expected = ['proj="PROJ_ID0"', 'proj="Cool Project!" (PROJ_ID1)']
        actual = formatter.proj_formatter.format_by_list_object(a)
        self.assertEqual(expected, actual)

    def test_format_exp(self):
        no_name_exp = MockWithNameResourceClasses("EXP_ID0")
        exp = MockWithNameResourceClasses("EXP_ID1", name="Cool Experiment!")
        self.assertEqual('exp="EXP_ID0"', formatter.exp_formatter.format(no_name_exp))
        self.assertEqual(
            'exp="Cool Experiment!" (EXP_ID1)', formatter.exp_formatter.format(exp)
        )

    def test_format_exp_list(self):
        a = [MockWithNameResourceClasses("EXP_ID{}".format(i)) for i in range(2)]
        a[1].data.name = "Cool Experiment!"
        expected = ['exp="EXP_ID0"', 'exp="Cool Experiment!" (EXP_ID1)']
        actual = formatter.exp_formatter.format_by_list_object(a)
        self.assertEqual(expected, actual)

    def test_format_exp_run(self):
        exp_run = MockWithID("RUN_ID")
        self.assertEqual('run="RUN_ID"', formatter.run_formatter.format(exp_run))

    def test_format_exp_run_list(self):
        a = [MockWithID("RUN_ID{}".format(i)) for i in range(2)]
        expected = ['run="RUN_ID0"', 'run="RUN_ID1"']
        actual = formatter.run_formatter.format_by_list_object(a)
        self.assertEqual(expected, actual)

    def test_exp_run_created_lines(self):
        exp_run = MockWithID("RUN_ID")
        exp_run.status = 1
        expected = ['run="RUN_ID"', 'status="QUEUED"']
        actual = formatter.run_formatter.run_created_lines(exp_run)
        self.assertEqual(expected, actual)

    def test_exp_run_checked_lines(self):
        exp_run = MockWithID("RUN_ID")
        exp_run.status = 1
        exp_run.algorithm = 6
        exp_run.hardware = 6
        exp_run.result = "ERROR"
        exp_run.error_message = ""

        expected = [
            'run="RUN_ID"',
            'status="QUEUED"',
            'algorithm="TOY"',
            'compute="LARGE_GPU"',
            'result="ERROR"',
            #'' signifies no error message expected as status is QUEUED
            # if error expected, this field should show:
            # error_message="Example error message."
            "",
        ]
        actual = formatter.run_formatter.run_checked_lines(exp_run)
        self.assertEqual(expected, actual)



from tq42.utils.file_handling import read_file
from tq42.utils.constants import (
    insufficient_permission_errors_file,
    invalid_arguments_error_file,
    no_default_error_file,
    unauthenticated_error_file,
)


class TQ42APIError(Exception):
    """Base class for custom exceptions in TQ42API."""

    pass


class ExceedRetriesError(TQ42APIError):
    def __init__(self, tries: int):
        self.tries = tries

    def __str__(self):
        return "Polling exceeded. Number of retries: {}".format(self.tries)


class NoDefaultError(TQ42APIError):
    def __init__(self, command: str):
        self.command = command

    def __str__(self):
        return read_file(no_default_error_file).format(self.command)


class InvalidArgumentError(TQ42APIError):
    def __init__(self, command: str, details: str):
        self.command = command
        self.details = details

    def __str__(self):
        return read_file(invalid_arguments_error_file).format(
            self.command, self.details
        )


class UnauthenticatedError(TQ42APIError):
    def __str__(self):
        return read_file(unauthenticated_error_file)


class PermissionDeniedError(TQ42APIError):
    def __str__(self):
        return read_file(insufficient_permission_errors_file)


class InvalidInputCliError(TQ42APIError):
    def __init__(self, msg: str):
        self.msg = msg

    def __str__(self):
        return "Invalid input: " + self.msg


class NoMatchingAttributeError(TQ42APIError):
    def __init__(self, details: str):
        self.details = details


class ExperimentRunCancelError(TQ42APIError):
    def __str__(self):
        return "Cannot cancel a run that was already completed or cancelled."


class AuthenticationError(TQ42APIError):
    def __init__(self, message="Authentication error occurred."):
        self.message = message
        super().__init__(self.message)



import unittest
import os
from dataclasses import dataclass
from datetime import datetime
from typing import Any
from unittest import mock
from keyring.errors import InitError, NoKeyringError, PasswordSetError, KeyringLocked

from tq42.utils import dirs, file_handling, exceptions
from tq42.utils.misc import (
    get_id,
    get_algo_num,
    get_hardware_num,
    dynamic_create_exp_run_request,
    find_oneof_field_name,
    save_token,
    get_token,
)
from tq42.utils.token_manager import TokenManager
import json
from tq42.client import ConfigEnvironment, TQ42Client
from tq42.utils.exceptions import NoMatchingAttributeError
from tq42.algorithm import AlgorithmProto
from tq42.experiment_run import HardwareProto
from com.terraquantum.experiment.v1.experimentrun import (
    create_experiment_run_request_pb2 as create_exp_run,
)


class TestUtils(unittest.TestCase):
    def setUp(self):
        self.client = TQ42Client()

    def test_utils_get_id(self):
        filepath = dirs.full_path(dirs.test_data_dir(), "test_utils_get_id.txt")
        content = file_handling.read_file(filepath)
        res = get_id(content)
        self.assertEqual('"64a35a73-3da2-4f44-bce3-9e5a8eac6d30"', res)

    def test_get_matching_algorithm(self):
        self.assertEqual(AlgorithmProto.TOY, get_algo_num("toy"))
        self.assertEqual(AlgorithmProto.TOY, get_algo_num("ToY"))
        self.assertEqual(AlgorithmProto.TOY, get_algo_num("TOY"))

        self.assertEqual(1, get_algo_num(1))

        self.assertEqual(AlgorithmProto.TOY, get_algo_num(AlgorithmProto.TOY))
        self.assertEqual(
            AlgorithmProto.TOY,
            get_algo_num(AlgorithmProto.Name(AlgorithmProto.TOY)),
        )

    def test_no_matching_algorithm(self):
        self.assertRaises(
            exceptions.NoMatchingAttributeError,
            get_algo_num,
            "no-way-this-is-an-algo",
        )
        self.assertRaises(exceptions.NoMatchingAttributeError, get_algo_num, "")

    def test_get_matching_hardware(self):
        self.assertEqual(HardwareProto.SMALL, get_hardware_num("small"))
        self.assertEqual(HardwareProto.SMALL, get_hardware_num("SmAlL"))
        self.assertEqual(HardwareProto.SMALL, get_hardware_num("SMALL"))

        self.assertEqual(1, get_hardware_num(1))

        self.assertEqual(
            HardwareProto.SMALL,
            get_hardware_num(HardwareProto.SMALL),
        )
        self.assertEqual(
            HardwareProto.SMALL,
            get_hardware_num(HardwareProto.Name(HardwareProto.SMALL)),
        )

    def test_no_matching_hardware(self):
        self.assertRaises(
            exceptions.NoMatchingAttributeError,
            get_hardware_num,
            "no-way-this-is-a-hardware-config",
        )
        self.assertRaises(exceptions.NoMatchingAttributeError, get_hardware_num, "")

    @mock.patch("requests.post")
    def test_renew_expiring_token(self, post_mock):
        @dataclass
        class MockResponse:
            json_data: Any
            status_code: Any

            def json(self):
                return self.json_data

        data = {
            "access_token": "fake_access_token",
            "id_token": "fake_id_token",
            "scope": "fake_scopes",
            "expires_in": 86400,
            "token_type": "Bearer",
        }

        post_mock.return_value = MockResponse(json_data=data, status_code=200)

        token_timestamp = "2022-07-21 13:48:26.580403"
        token_timestamp = datetime.strptime(token_timestamp, "%Y-%m-%d %H:%M:%S.%f")
        # force expire timestamp
        file_handling.write_to_file(
            self.client.timestamp_file_path, str(token_timestamp)
        )

        # load config file
        alt_config_file = dirs.text_files_dir("config.json")
        with open(alt_config_file, encoding="utf-8") as f:
            config_data = json.load(f)

        environment = ConfigEnvironment(
            config_data["base_url"],
            config_data["client_id"],
            config_data["scope"],
        )

        token_manager = TokenManager(environment, None)

        success = token_manager.renew_expring_token()
        # old token timestamp should renew token
        self.assertTrue(success)

        # renew should return false as timestamp has been updated already
        success = token_manager.renew_expring_token()
        # new token timestamp should NOT renew token
        self.assertFalse(success)

    def test_dynamic_request_valid_params(self):
        params = {
            "parameters": {"n": 4, "r": 1.1, "msg": "this is optimus prime"},
            "inputs": {},
        }
        result = dynamic_create_exp_run_request(
            parameters=params,
            algo=AlgorithmProto.TOY,
            exp_id="random-uuid",
            hardware=HardwareProto.SMALL,
        )
        self.assertIsNot(result.SerializeToString(), b"")

    def test_dynamic_request_invalid_params_field_name(self):
        params = {
            "parameters": {
                "msg": "this is bumblebee",
                "unknown field name": "this field should not be here",
            },
            "inputs": {},
        }
        with self.assertRaises(NoMatchingAttributeError) as context:
            dynamic_create_exp_run_request(
                parameters=params,
                algo=AlgorithmProto.TOY,
                exp_id="random-uuid",
                hardware=HardwareProto.SMALL,
            )

        self.assertEqual(
            str(context.exception.details),
            'Protocol message ToyParametersProto has no "unknown field name" field.',
        )

    def test_dynamic_request_invalid_params_grid(self):
        params = {
            "parameters": {
                "dimensionality": 4,
                "iteration_number": 1,
                "maximal_rank": 4,
                "points_number": 1,
                "quantization": True,
                "tolerance": 0.001,
                "lower_limits": [0, 0.2, 0.5, 0.2],
                "upper_limits": [30, 0.5, 1.5, 0.6],
                "grd": [5, 5, 5, 5],
                "objectivefunction": "http://34.141.232.153:8000/ymixer/eval",
            },
            "inputs": {},
        }

        with self.assertRaises(NoMatchingAttributeError) as context:
            dynamic_create_exp_run_request(
                parameters=params,
                algo=AlgorithmProto.TETRA_OPT,
                exp_id="random-uuid",
                hardware=HardwareProto.SMALL,
            )

        self.assertEqual(
            str(context.exception.details),
            'Protocol message TetraOptParametersProto has no "grd" field.',
        )

    def test_dynamic_request_params(self):
        parameters = {
            "parameters": {
                "dimensionality": 4,
                "iteration_number": 1,
                "maximal_rank": 4,
                "points_number": 1,
                "quantization": True,
                "tolerance": 0.001,
                "lower_limits": [0, 0.2, 0.5, 0.2],
                "upper_limits": [30, 0.5, 1.5, 0.6],
                "grid": [5, 5, 5, 5],
                "objective_function": "http://34.141.232.153:8000/ymixer/eval",
            },
            "inputs": {},
        }

        result = dynamic_create_exp_run_request(
            parameters=parameters,
            algo=AlgorithmProto.TETRA_OPT,
            exp_id="random-uuid",
            hardware=HardwareProto.SMALL,
        )

        expected = create_exp_run.CreateExperimentRunRequest(
            request_id=None,
            algorithm=AlgorithmProto.TETRA_OPT,
            experiment_id="random-uuid",
            hardware=HardwareProto.SMALL,
            tetra_opt_metadata=parameters,
        )

        self.assertEqual(str(result), str(expected))

    def test_config_file_path(self):
        alternative_json_path = dirs.testdata("config.json")
        tq42 = TQ42Client(alternative_json_path)
        self.assertEqual(tq42.config_file, alternative_json_path)

    def test_find_oneof_field_name(self):
        res = find_oneof_field_name("ToyMetadataProto")
        self.assertEqual(res, "toy_metadata")

    def test_save_get_token_with_keyring_enabled(self):
        # keyring is working on Mac Sonoma 14.4 and Windows 11
        token_file_path = os.path.join(dirs.testdata(), "keyring_test.json")
        save_token(
            service_name="tq42_access_token",
            backup_save_path=token_file_path,
            token="test_token",
        )
        token = get_token(
            service_name="tq42_access_token", backup_save_path=token_file_path
        )
        self.assertEqual(token, "test_token")

    @mock.patch("keyring.set_password")
    @mock.patch("keyring.get_password")
    def test_save_get_token_has_InitError(self, mock_set_password, mock_get_password):
        # keyring not working on Ubuntu 20.04.6LTS and causes InitError exception
        token_file_path = os.path.join(dirs.testdata(), "keyring_test.json")
        mock_set_password.side_effect = InitError()
        mock_get_password.side_effect = InitError()
        save_token(
            service_name="tq42_access_token",
            backup_save_path=token_file_path,
            token="test_token",
        )
        token = get_token(
            service_name="tq42_access_token", backup_save_path=token_file_path
        )
        os.remove(token_file_path)
        self.assertEqual(token, "test_token")

    @mock.patch("keyring.set_password")
    @mock.patch("keyring.get_password")
    def test_save_get_token_has_NoKeyringError(
        self, mock_set_password, mock_get_password
    ):
        # keyring not working on Debian and Redhat, NoKeyringError is raised
        token_file_path = os.path.join(dirs.testdata(), "keyring_test.json")
        mock_set_password.side_effect = NoKeyringError()
        mock_get_password.side_effect = NoKeyringError()
        save_token(
            service_name="tq42_access_token",
            backup_save_path=token_file_path,
            token="test_token",
        )
        token = get_token(
            service_name="tq42_access_token", backup_save_path=token_file_path
        )
        os.remove(token_file_path)
        self.assertEqual(token, "test_token")

    @mock.patch("keyring.get_password")
    @mock.patch("keyring.set_password")
    def test_should_return_token_even_if_keyring_locked_and_passwort_set_error_thrown(
        self, mock_set_password, mock_get_password
    ):
        token_file_path = os.path.join(dirs.testdata(), "keyring_test.json")
        mock_set_password.side_effect = PasswordSetError()
        mock_get_password.side_effect = KeyringLocked()

        save_token(
            service_name="tq42_access_token",
            backup_save_path=token_file_path,
            token="test_token",
        )

        token = get_token(
            service_name="tq42_access_token", backup_save_path=token_file_path
        )
        os.remove(token_file_path)
        self.assertEqual(token, "test_token")



import unittest

from tq42.functional_tests.functional_test_config import FunctionalCLITestConfig


class TestFunctionalInvokeCliOrganizationGroup(
    unittest.TestCase, FunctionalCLITestConfig
):
    def test_org_list(self):
        result = self.runner.invoke(self.cli_entry, ["org", "list"])
        assert result.exit_code == 0
        assert result.output is not None

    def test_org_set(self):
        result = self.runner.invoke(self.cli_entry, ["org", "set", self.org])
        assert result.exit_code == 0
        assert result.output is not None



import unittest

from tq42.cli import (
    experiment_group,
    experiment_run_group,
    organization_group,
    project_group,
)
from tq42 import algorithm, organization
import tq42.cli.utils.formatter as formatter
from tq42.utils import dirs, misc
from tq42.client import TQ42Client


class TestImports(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_tq42_imports(self):
        self.assertIsNotNone(algorithm)
        self.assertIsNotNone(organization)

    def test_utils_imports(self):
        self.assertIsNotNone(TQ42Client)
        self.assertIsNotNone(dirs)
        self.assertIsNotNone(misc)

    def test_cli_imports(self):
        self.assertIsNotNone(experiment_group)
        self.assertIsNotNone(experiment_run_group)
        self.assertIsNotNone(organization_group)
        self.assertIsNotNone(project_group)

    def test_formatter_imports(self):
        self.assertIsNotNone(formatter)



import click

from tq42.cli.utils.types import TQ42CliContext
from tq42.model import Model, list_all


@click.group("model")
def model_group() -> click.Group:
    """
    Group to manage models.

    https://docs.tq42.com/en/latest/Python_Developer_Guide/Working_with_Models.html
    """
    pass


@model_group.command("list")
@click.option(
    "--proj",
    "proj_id",
    required=True,
    type=str,
    help="Attribute the command to a particular project",
)
@click.pass_context
def list_models(ctx: TQ42CliContext, proj_id: str) -> None:
    """
    List all models within a project.

    e.g.: tq42 proj model list --proj 98ccb1d2-a3d0-48c8-b172-022f6db9be01

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Working_with_Models.html
    """
    click.echo(
        [model.data for model in list_all(client=ctx.obj.client, project_id=proj_id)]
    )


@model_group.command("get")
@click.argument("model_id", required=True, type=str)
@click.pass_context
def get(ctx: TQ42CliContext, model_id: str) -> None:
    """
    Get a specific model.

    e.g.: tq42 proj model get 98ccb1d2-a3d0-48c8-b172-022f6db9be01

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Working_with_Models.html
    """
    click.echo(Model(client=ctx.obj.client, id=model_id).data)



import click

from tq42.cli.utils.types import TQ42CliContext


@click.group("auth")
def auth_group() -> click.Group:
    """
    Group to manage authentication

    https://docs.tq42.com/en/latest/README.html#authentication
    """
    pass


@auth_group.command("login")
@click.pass_context
def login(ctx: TQ42CliContext) -> None:
    """
    Opens a browser window to confirm the MFA code, enter your TQ42 username and password to authenticate.
    To access TQ42 services with Python commands, you need a TQ42 account.
    When running TQ42 Python commands, your environment needs to have access to your TQ42 account credentials.

    https://docs.tq42.com/en/latest/README.html#authentication
    """
    ctx.obj.client.login()



import unittest

from tq42.functional_tests.functional_test_config import FunctionalCLITestConfig


class TestFunctionalCli(unittest.TestCase, FunctionalCLITestConfig):
    def test_proj_set(self):
        result = self.runner.invoke(self.cli_entry, ["proj", "set", self.proj])
        assert result is not None

    def test_proj_show(self):
        result = self.runner.invoke(self.cli_entry, ["proj", "show"])
        assert result.exit_code == 0
        assert "org=" in result.output

    def test_list_orgs(self):
        result = self.runner.invoke(self.cli_entry, ["org", "list"])
        assert "org=" in result.output

    def test_list_proj_by_org(self):
        result = self.runner.invoke(self.cli_entry, ["proj", "list", "--org", self.org])
        assert "proj=" in result.output

    def test_get_exprun(self):
        result = self.runner.invoke(
            self.cli_entry, ["exp", "run", "check", self.exp_run]
        )
        assert "status" in result.output



from tq42.utils import dirs, file_handling
from tq42.utils.misc import get_token, save_token
from datetime import datetime
import requests


class TokenManager:
    def __init__(self, environment, alt_config_folder):
        self.alt_config_folder = alt_config_folder
        self.environment = environment

    @property
    def token_file_path(self):
        token_folder = dirs.create_or_get_config_dir(self.alt_config_folder)
        return dirs.full_path(token_folder, "token.json")

    @property
    def timestamp_file_path(self):
        token_folder = dirs.create_or_get_config_dir(self.alt_config_folder)
        return dirs.full_path(token_folder, "timestamp.json")

    @property
    def refresh_token_file_path(self):
        token_folder = dirs.create_or_get_config_dir(self.alt_config_folder)
        return dirs.full_path(token_folder, "refresh_token.json")

    def renew_expring_token(self):
        token_timestamp = file_handling.read_file(self.timestamp_file_path)
        if token_timestamp == "":
            return False
        token_timestamp = datetime.strptime(token_timestamp, "%Y-%m-%d %H:%M:%S.%f")
        diff = datetime.now() - token_timestamp
        # get new access token if token timestamp more than 23 hours(82800 seconds)
        # refresh token is valid for 30 days
        renew_limit = 82800
        if diff.total_seconds() > renew_limit:
            self.request_new_access_token()
            return True

        return False

    def request_new_access_token(self):
        refresh_token = get_token(
            service_name="tq42_refresh_token",
            backup_save_path=self.refresh_token_file_path,
        )
        data = self.environment.refresh_token_data(refresh_token)
        response = requests.post(
            self.environment.auth_url_token,
            data=data,
            headers=self.environment.headers,
        )
        json_response = response.json()
        if "access_token" in json_response:
            access_token = json_response["access_token"]
            save_token(
                service_name="tq42_access_token",
                backup_save_path=self.token_file_path,
                token=access_token,
            )
            current_datetime = datetime.now()
            file_handling.write_to_file(self.timestamp_file_path, current_datetime)
